"""
The `sql` module provides helper functions for working with SQL tables in Microsoft Fabric 
This module includes functions for creating temporary views for all underlying SQL tables in a SQL query. 
This is particularly useful when working with large datasets or complex queries.

 - find_all_tables_in_sql: Finds all source tables used in a SQL query.
 - create_sql_views_from_sql: Creates a temporary view for each table in a SQL query.
 - add_default_where_clause: Adds a default where clause to a configuration file.
 - get_table_name_and_alias: Retrieves the table name from the table name string and the alias if exists.

Here's an example of how to use this module:

```python
from evidi_fabric.sql import create_sql_views_from_sql
sql = "SELECT sales_amount FROM transactions"
lakehouse = "Decide_DK_Silver"
create_sql_views_from_sql(sql=sql, lakehouse=lakehouse)
"""

import re
from uuid import UUID
from evidi_fabric.fs import get_table_path

try:
    spark
except NameError:
    from evidi_fabric.spark import get_or_create_spark
    spark = get_or_create_spark()


def find_all_tables_in_sql(sql_query:str) -> list[str]:
    """
    From any sql_query, returns a list of all dependent tables.
    
    e.g. 
    sql_query="SELECT * FROM A" -> ["A"]
    sql_query="WITH CTE (SELECT * FROM A)\n SELECT * FROM CTE" -> ["A","CTE"]
    sql_query="WITH CTE (SELECT * FROM A)\n SELECT * FROM CTE INNER JOIN B ON A.key = B.key" -> ["A","CTE","B"]
    
    """
    pattern = re.compile(r'\b(?:FROM|JOIN)\s+([a-zA-Z_]\w*)\b', re.IGNORECASE)
    matches = pattern.findall(sql_query)
    return matches

def create_sql_views_from_sql(sql:str, lakehouse:str | UUID | None = None) -> None:
    """
    When reading a table from an sql query, i.e. spark.sql(table_sql), 
    you cannot reference the absolute path to the source.

    To overcome this, this function 
    1) finds all source tables used in the sql_query,
    2) finds their absolute paths
    3) create a temporary view of the data

    This view can then be accessed from sql, but refere to the table name 
    without specifying the schema

    Example:
        table_sql = "SELECT sales_amount FROM transactions"
        lakehouse = "Decide_DK_Silver"

    then a view is created called transactions, 
    so that spark.sql(table_sql) can be executed pointing to the 
    absolute path of transactions
    """
    table_names=find_all_tables_in_sql(sql)
    for table_name in table_names:
        table_path=get_table_path(table_name=table_name, lakehouse=lakehouse)
        df_temp=spark.read.load(table_path)
        df_temp.createOrReplaceTempView(table_name)

def add_default_where_clause(config:dict) -> dict:
    """
    In the silver layer every destination table and its corresponding company
    must have a where clause. In case that it is 
    1) not set or
    2) set to None or
    3) set to empty string
    
    then it is replaced with 1=1.
    """
    for destination_table in config.keys():
        for company in config[destination_table]["companies"].keys():
            if config[destination_table]["companies"][company].get("where_clause") in [None, ""]:
                config[destination_table]["companies"][company]["where_clause"]="1=1"
    return config


def get_table_name_and_alias(table_name_raw:str) -> tuple[str, str]:
    """
    Retrieves the table name from the table name string and the alias if exists.
    If alias is not set, then the table name is used as the alias.

    Examples:
        - 'A__CUSTABLE' -> ('A__CUSTABLE', 'A__CUSTABLE')
        - 'A__CUSTABLE as A' -> ('A__CUSTABLE', 'A')
        - 'A__CUSTABLE AS   B ' -> ('A__CUSTABLE', 'B')
    """
    table_name_raw=table_name_raw.strip()

    if " as " in table_name_raw.lower():
        table_name_raw=re.sub(" +", " ", table_name_raw)
        index=table_name_raw.lower().find(" as ")
        table_name=table_name_raw[:index]
        table_alias=table_name_raw[index+4:]
    
    else:
        table_name=table_name_raw
        table_alias=table_name_raw
    return table_name, table_alias

if __name__=="__main__":
    table_sql = "WITH CTE (SELECT * FROM A)\n SELECT * FROM CTE INNER JOIN B ON A.key"
    tables = find_all_tables_in_sql(table_sql)
    print(tables)

    sql = "SELECT sales_amount FROM transactions"
    lakehouse = "Decide_DK_Silver"
    create_sql_views_from_sql(sql, lakehouse)

    


"""
This module contains functions for cleaning dataframes.

 - clean_df: Cleans a dataframe according to the configuration file.

Here's an example of how to use this module:
```python
from evidi_fabric.clean import clean_df 

table = "FO_A__SALESLINES"
clean_df(df=df, table_name=table, config=CONFIG_SILVER)
```
"""


from pyspark.sql import DataFrame, functions as f
from evidi_fabric.fs import get_table_path, resolve_workspace_name, resolve_lakehouse_name

try:
    spark
except NameError:
    from evidi_fabric.spark import get_or_create_spark
    spark = get_or_create_spark()


def clean_df(df:DataFrame, table_name:str, config:dict, lakehouse:str=None, workspace:str=None) -> DataFrame:
    df = df.dropDuplicates()
    df = _add_currency_converted_cols(df=df, 
                                     table_name=table_name, 
                                     config=config, 
                                     lakehouse=lakehouse, 
                                     workspace=workspace)    
    return df

def _add_currency_converted_cols(df:DataFrame, table_name:dict[str,str], config:dict, lakehouse:str=None, workspace:str=None) -> DataFrame:
    """
    Adds three column to the existing dataframe for each col_name in the "convert_currency_cols" element of the CONFIG_SILVER

    Input:
        df: DataFrame
        directory_info: {"directory": <table_name>,
                         "path": <absolute path to table>}

    Output:
        df: The same dataframe, but with added columns, for each of the above mentioned element:

        Added columns:
            f"{col_name}__Rate": Conversion rate
            f"{col_name}__CONV": The converted price
            f"{col_name}__CONV_ERR": 1 if the price could not be converted. Zero otherwise
    """
    if workspace == None:
        workspace = resolve_workspace_name()
    if lakehouse == None:
        lakehouse = resolve_lakehouse_name()

    df_merged=df
    if table_name in config.keys():
        config_table=config[table_name]
        if "convert_currency" in config_table:
            config_currency=config_table["convert_currency"]
            currency_table_name=config_currency["depends_on"]
            currency_table_path=get_table_path(currency_table_name, lakehouse)
            df_currency=spark.read.load(currency_table_path)
            for convert_currency_col_info in config_currency["convert_currency_cols"]:
                date_col:str=convert_currency_col_info["date_col"]
                col_name:str=convert_currency_col_info["col"]
                if "dataareaid_col" in convert_currency_col_info.keys():
                    dataareaid_col=convert_currency_col_info["dataareaid_col"]
                    join_clause:tuple=(f.col(f"df.{date_col}") == df_currency.Date)&\
                                      (f.upper(f.col(f"df.{dataareaid_col}")) == df_currency.DataAreaId)
                elif "currency_col" in convert_currency_col_info.keys():
                    currency_col=convert_currency_col_info["currency_col"]
                    join_clause:tuple=(f.col(f"df.{date_col}").cast("date") == df_currency.Date)&\
                                      (f.upper(f.col(f"df.{currency_col}")) == f.upper(df_currency.CurrencyCode))
                #TODO: join_clause must be wrong, but is as Flemming defined it
                #join_clause:tuple=(datetime(2024,1,1) == df_currency.Date)&(f.upper(df_merged.DATAAREAID) == df_currency.DataAreaId)
                df_merged=df_merged.alias("df") \
                            .join(df_currency.alias("currency"), join_clause, how="left") \
                            .select(f"df.*", f.col("currency.Rate") \
                                              .alias(f"{col_name}__RATE"))
                df_merged=df_merged.withColumn(f"{col_name}__CONV", f.col(col_name) * f.col(f"{col_name}__Rate"))
                df_merged=df_merged.withColumn(f"{col_name}__CONV_ERR", f.when(f.col(f"{col_name}__CONV").isNull(),1) \
                                                                         .otherwise(0))
    return df_merged

if __name__ == "__main__":
    pass


"""
The `utils` module provides utility functions for handling and manipulating data structures.

 - extend_list_in_dict: Extends list element of a dictionary.
 - merge_configs: Merges two dictionaries together.
 - get_config: Retrieves and merge the config_base and config_custom config files to a single config dictionary.

Here's an example of how to use this module:

```python
from evidi_fabric.utils import extend_list_in_dict, merge_configs

# Merge configurations
base_config = {'key1': ['value1']}
custom_config = {'key1': ['value2']}
merged_config = merge_configs(base_config, custom_config)
"""

from evidi_fabric.sql import add_default_where_clause

def extend_list_in_dict(dct, key, value:list) -> dict:
    """
    Extending list element of a dictionary. 
    If key does not exists, then it is created.
    """
    if key in dct.keys():
        dct[key].extend(value)
    else:
        dct[key]=value
    return dct


def merge_configs(base:dict, custom:dict) -> dict:
    """
    Find matching keys from two dictionaries and append the lists together if value is list.
    """
    result = base.copy()
    for key, value in custom.items():
        if isinstance(value, dict) and key in base:
            result[key] = merge_configs(base[key], value)
        elif isinstance(value, list):
            # Append additional columns to the existing 'columns' list
            result[key] = base.get(key, []) + value
        else:
            result[key] = value
    return result


def get_config(config_base, config_custom) -> dict:
    """
    Retrieves and merge the config_base and config_custom config files to a single config dictionary. 
    """
    config_raw = merge_configs(config_base, config_custom)
    try:
        config = add_default_where_clause(config_raw)
    except:
        #Could not add default where clause
        config = config_raw
        pass

    return config


from evidi_fabric.sql import find_all_tables_in_sql

def extract_depends_on(table_infos, source_table_names:list[str]):
    """
	Looks through the table_infos and returns a list of all tables that the table depends on
    and are in the source_table_names.

    This function is recursive and will look through all nested dictionaries.
	"""

    depends_on_list = []
    if "depends_on" in table_infos:
        depends_on = table_infos["depends_on"]
        if isinstance(depends_on, list):
            for item in depends_on:
                if isinstance(item, dict) and "table" in item:
                    if item["table"] in source_table_names:
                        depends_on_list.append(item["table"])
                if isinstance(item, dict) and "table_sql" in item:
                    tables = find_all_tables_in_sql(sql_query=item["table_sql"])
                    tables = [table for table in tables if table in source_table_names]
                    depends_on_list.extend(tables)
                elif isinstance(item, str):
                    if item in source_table_names:
                        depends_on_list.append(item)
        elif isinstance(depends_on, str):
            if depends_on in source_table_names:
                depends_on_list.append(depends_on)

    for _, value in table_infos.items():
        if isinstance(value, dict):
            depends_on_list.extend(extract_depends_on(table_infos=value, 
                                                      source_table_names=source_table_names))

    return depends_on_list


def get_dependencies(config, source_table_names:list[str]) -> dict[str, list[str]]:
    """
    Returns a dictionary of all dependencies for each table in the config 
    where the dependencies are in the source_table_names.
    """
    relevant_dependencies = {}
    for table_name, table_infos in config.items():
        depends_on_list = extract_depends_on(table_infos=table_infos, 
                                             source_table_names=source_table_names)
        if depends_on_list:
            relevant_dependencies[table_name] = depends_on_list
    return relevant_dependencies

def _get_DAG_from_dependencies(silver_tables_with_new_data:list, 
                               silver_dependencies:dict[str,list[str]], 
                               gold_dependencies:dict[str,list[str]], 
                               max_concurrency:int=20,
                               timeout_in_secs:int=3600) -> dict:
    """
    Returns a Directed Acyclic Graph (DAG) from the silver and gold dependencies.

    First, silver tables that depends on other silver tables, e.g. currency tables, are added to the DAG.
    Then, the remaining silver tables added to the DAG, with there eventual dependencies.
    Finally, the gold tables are added to the DAG, with there eventual dependencies.

    Args:
        silver_tables_with_new_data (list): List of silver tables with new data.
        silver_dependencies (dict): Dictionary mapping silver tables to their dependencies.
        gold_dependencies (dict): Dictionary mapping gold tables to their dependencies.
        max_concurrency (int, optional): Maximum number of concurrent activities in the DAG. Defaults to 20.
        timeout_in_secs (int, optional): Timeout for the entire pipeline, in seconds. Defaults to 3600.

    """

    # Convert to DAG
    DAG = {"activities": []}

    # Add dependent silver activities to DAG
    silver_dependencies_tables=list(set([dependency for dependencies in silver_dependencies.values() for dependency in dependencies]))
    for dependency in silver_dependencies_tables:
        activity = {
            "name": dependency,
            "path": "STD",
            "args": {"tables":dependency, "useRootDefaultLakehouse": True},
            
        }
        DAG["activities"].append(activity)

    # Add remaining silver activities to DAG
    silver_tables_with_new_data = [table for table in silver_tables_with_new_data if table not in silver_dependencies_tables]
    for table in silver_tables_with_new_data:
        if table in silver_dependencies.keys():
            dependencies = silver_dependencies[table]
        else:
            dependencies = []
        activity = {
            "name": table,
            "path": "STD__FO_A",
            "args": {"tables":table, "useRootDefaultLakehouse": True},
            "dependencies": dependencies,
        }
        DAG["activities"].append(activity)

    # Add gold activities to DAG
    for table, dependency in gold_dependencies.items():
        activity = {
            "name": table,
            "path": "TRANS",  
            "args": {"destination_table_name":table, "useRootDefaultLakehouse": True},
            "dependencies": dependency, 
        }
        DAG["activities"].append(activity)
        DAG["concurrency"] = max_concurrency # max 5 notebooks in parallel
        #see more: https://www.linkedin.com/pulse/micosoft-fabric-run-multiple-notebooks-parallel-dependencies-uerrc/?trk=organization_guest_main-feed-card_feed-article-content
        DAG["timeoutInSeconds"] = timeout_in_secs, # max 1 hour for the entire pipeline
    return DAG


def get_DAG(config_silver:dict[str,list[str]], 
            config_gold:dict[str,list[str]], 
            tables:list[str],
            max_concurrency:int=20,
            timeout_in_secs:int=3600) -> dict:
    """
    From a list of tables and config files that entails dependency infos, retrieve a DAG
    """
    silver_dependencies = get_dependencies(config=config_silver, 
                                           source_table_names=tables)
    
    silver_tables_with_new_data=list(set(list(silver_dependencies.keys())+tables))

    gold_dependencies = get_dependencies(config=config_gold, 
                                         source_table_names=silver_tables_with_new_data)
    
    DAG = _get_DAG_from_dependencies(silver_tables_with_new_data=silver_tables_with_new_data, 
                                     silver_dependencies=silver_dependencies, 
                                     gold_dependencies=gold_dependencies, 
                                     max_concurrency=max_concurrency, 
                                     timeout_in_secs=timeout_in_secs)
    return DAG


if __name__=="__main__":
    import json
    config_silver = {}
    config_gold = {"DimCompany":
	{
		"load_type": "FULL",
		 "primary_keys" : [
			"CompanyKey"
		 ],
		 "companies":{
			"A":{
			 "depends_on": [
				{
					"table": "FO_A__DATAAREA"
				}
			 ],
			 "columns": [
				"LOWER(CONCAT('A#',FO_A__DATAAREA.ID)) AS CompanyKey",
				"CONCAT(FO_A__DATAAREA.ID, ' - ', FO_A__DATAAREA.NAME) AS Company",
				"FO_A__DATAAREA.ID AS CompanyId",
				"FO_A__DATAAREA.NAME AS CompanyName",
			  ],
			 "where_clause":"FO_A__DATAAREA.PARTITION=5637144576",
			}
		}
	}}
    tables = ["FO_A__DATAAREA"]
    DAG = get_DAG(config_silver, config_gold, tables)
    print(json.dumps(DAG, indent=2))
    


"""
This package provides a collection of modules for data processing and manipulation.

Submodules:
- clean: Contains functions for cleaning dataframes.
- fs: Provides functions for file system operations.
- sql: Includes helper functions for working with SQL.
- utils: Provides utility functions for handling and manipulating data structures.

You can import these modules like so:

```python
from my_package import clean, fs, sql, utils
"""


import logging
from pyspark.conf import SparkConf
from pyspark.sql.session import SparkSession

from evidi_fabric.config import settings

logging.getLogger(__name__)


def get_or_create_spark() -> SparkSession:
    """Gets a spark session and returns it unless it already exists.

    This function is used to start a spark session and return it. It is used
    to avoid creating multiple spark sessions. In Databricks, this function
    simply returns the existing spark session.

    Returns:
        SparkSession: Spark session
    """
    logging.info("Getting or creating spark session")

    conf = SparkConf()
    try:
        conf.set("spark.jars.packages", ",".join(settings.packages))
    except AttributeError:
        pass

    try:
        conf.set("spark.driver.memory", settings.spark.driver.memory)
    except AttributeError:
        conf.set("spark.driver.memory", "8g")
    conf.set("spark.driver.maxResultSize", "4g")
    conf.set("spark.sql.execution.arrow.pyspark.enabled", "true")
    conf.set("spark.sql.debug.maxToStringFields", "1000")  # to avoid warning
    conf.set("spark.sql.extensions", "io.delta.sql.DeltaSparkSessionExtension")
    conf.set("spark.sql.catalog.spark_catalog", "org.apache.spark.sql.delta.catalog.DeltaCatalog")
    
    try:
        conf.set('spark.sql.session.timeZone', settings.spark.sql.session.timeZone)
    except:
        conf.set('spark.sql.session.timeZone', 'UTC')

    spark = SparkSession.builder.config(conf=conf).getOrCreate()
    return spark


if __name__ == "__main__":
    spark = get_or_create_spark()
    print('Done')



"""
The `fs` module provides functions for file system operations.

This includes functions for getting table paths, resolving workspace names, and resolving lakehouse names. One of the key functions in this module is `get_table_from_file_or_folder`, which retrieves the table from a given file or folder.

 - get_tables_from_files_or_folders: Retrieves the table names from a list of files or folders.
 - get_table_from_file_or_folder: Retrieves the table name from a file or folder path.
 - get_file_path: Retrieves a fully qualified file path.
 - get_table_path: Retrieves a fully qualified table path.
 - get_lakehouse_from_path: Retrieves the lakehouse name from a path.
 - resolve_workspace_id: Retrieves the workspace id from the workspace name.
 - resolve_workspace_name: Retrieves the workspace name from the workspace id.
 - resolve_lakehouse_id: Retrieves the lakehouse id from the lakehouse name.
 - resolve_lakehouse_name: Retrieves the lakehouse name from the lakehouse id.
 - get_lakehouse_tables_info: Retrieves the table names and related infos from the lakehouse.

Here's an example of how to use `get_table_from_file_or_folder`:

```python
from evidi_fabri.fs import get_table_from_file_or_folder

table = get_table_from_file_or_folder('path/to/file_or_folder')
```
"""
import pandas as pd
import sempy.fabric as fabric
from pyspark.sql.session import SparkSession
from uuid import UUID

from evidi_fabric.spark import get_or_create_spark

def _is_valid_uuid(val):
    try:
        UUID(str(val))
        return True
    except ValueError:
        return False
    
def get_tables_from_files_or_folders(files_or_folders:list[str]):
    """
    Either all files from bronze is passed from bronze til silver, or only a single file.
    This function will return a list of the table names
    """
    return [get_table_from_file_or_folder(file_or_folder) for file_or_folder in files_or_folders]

def get_table_from_file_or_folder(file_or_folder:str):
    """
    Either all files from bronze is passed from bronze til silver, or only a single file.
    This function will return the table name, by assuming the name of the silver table is 
    the name of the folder where the file is located. 
    
    This function looks after a "." in the file_or_folder string. If there is a dot, then
    it must be a full path if it's not, then its assumed to point to th

    E.g. 
        file_or_folder='table_name/20240101_filname.parquet' -> 'table_name'
        file_or_folder='table_name' -> 'table_name'
        file_or_folder='parent_folder/table_name' -> 'table_name'
        file_or_folder='filename.parquet' -> ValueError #Path must be specified

    """
    if "." in file_or_folder:
        if "/" in file_or_folder:
            return file_or_folder.split("/")[-2]
        else:
            raise ValueError("At least the relative path must be specified")
    else:
        if "/" in file_or_folder:
            return file_or_folder.split("/")[-1]
        else:
            return file_or_folder

def get_file_path(directory:str=None, lakehouse:str | UUID | None = None, workspace:str | UUID | None = None) -> str:
    """
    Retrieves a fully qualified file path.
    Note: 

    Params:
     - workspace: Name or id of workspace (e.g. Decide_DK_DEV, 7dae828c-f98e-4e08-aaae-d93b2774c74b, None)
        * If not provided, the current workspace is used
     - lakehouse: Name or id of the lakehouse (e.g. Decide_DK_Silver, 9be79118-d950-4662-8053-b393d021ab0f, None)
        * If not provided, the current lakehouse is used
     - directory: Name of the directory (e.g. Bronze)

    Output:
     - path (e.g. abfss://Decide_DK_DEV@onelake.dfs.fabric.microsoft.com/Decide_DK_Silver.Lakehouse/Files/Bronze/)
    """
    workspace_name = resolve_workspace_name(workspace)
    lakehouse_name = resolve_lakehouse_name(lakehouse)
    return f"abfss://{workspace_name}@onelake.dfs.fabric.microsoft.com/{lakehouse_name}.Lakehouse/Files/{directory}"


def get_table_path(table_name:str, lakehouse:str | UUID | None = None, workspace:str | UUID | None = None) -> str:
    """
    Retrieves a fully qualified table path.

    Params:
     - table_name: Name of the table (e.g. A__CUSTGROUP)
     - workspace: Name or id of workspace (e.g. Decide_DK_DEV, 7dae828c-f98e-4e08-aaae-d93b2774c74b, None)
        * If not provided, the current workspace is used
     - lakehouse: Name or id of the lakehouse (e.g. Decide_DK_Silver, 9be79118-d950-4662-8053-b393d021ab0f, None)
        * If not provided, the current lakehouse is used

    Output:
     - path (e.g. abfss://Decide_DK_DEV@onelake.dfs.fabric.microsoft.com/Decide_DK_Silver.Lakehouse/Tables/A__CUSTGROUP/
    """
    workspace_name = resolve_workspace_name(workspace)
    lakehouse_name = resolve_lakehouse_name(lakehouse)
    
    return f"abfss://{workspace_name}@onelake.dfs.fabric.microsoft.com/{lakehouse_name}.Lakehouse/Tables/{table_name}" 

            
def get_lakehouse_from_path(path:str) -> str:
    """
    Retrieves the lakehouse name from a path.

    Example:
    path = "abfss://Decide_DK_DEV@onelake.dfs.fabric.microsoft.com/Decide_DK_Bronze.Lakehouse/Files/Bronze/"
    lakehouse = "Decide_DK_Bronze"
    """
    if "/Files" in path:
        lakehouse = path.split('onelake.dfs.fabric.microsoft.com/')[-1].split('/Files')[0].replace('.Lakehouse','')
    elif "/Tables" in path:
        lakehouse = path.split('onelake.dfs.fabric.microsoft.com/')[-1].split('/Tables')[0].replace('.Lakehouse','')
    else:
        raise ValueError("The path is not a valid lakehouse path.")
    return lakehouse

def resolve_workspace_id(workspace:str | UUID | None = None) -> UUID:
    """
    From the workspace, that can be either the name, guid or None, resolve the workspace id.
    If no workspace is provided, the current workspace is used.
    """
    if workspace == None:        
        workspace_id = fabric.get_workspace_id()
    elif _is_valid_uuid(workspace):
        workspace_id = workspace
    else:
        workspace_id = fabric.resolve_workspace_id(workspace)

    return workspace_id

def resolve_workspace_name(workspace:str | UUID | None = None) -> str:
    """
    From the workspace, that can be either the name, guid or None, retrieve the workspace name.
    If no workspace id provided, the current workspace is used.
    """
    if workspace == None:
        workspace_id = resolve_workspace_id()
        workspace_name = fabric.resolve_workspace_name(workspace_id)

    elif not _is_valid_uuid(workspace):
        workspace_name=workspace
    
    else:
        workspace_id = workspace
        workspace_name = fabric.resolve_workspace_name(workspace_id)
    
    return workspace_name


def resolve_lakehouse_id(lakehouse:str | UUID | None = None) -> UUID:
    """
    From the lakehouse name, that can be either the name, guid or None, retrieves the lakehouse id.
    If no lakehouse name is provided, the current lakehouse is used.
    """
    if lakehouse == None:
        lakehouse_id = fabric.get_lakehouse_id()

    elif _is_valid_uuid(lakehouse):
        lakehouse_id = lakehouse

    else:
        df_items = fabric.list_items("Lakehouse")
        df_items = df_items[df_items['Display Name'] == lakehouse]
        lakehouse_id = df_items['Id'].iloc[0]

    return lakehouse_id

def resolve_lakehouse_name(lakehouse:str | UUID | None = None) -> str:
    """
    From the lakehouse, that can be either the name, guid or None, retrieves the lakehouse name.
    If no lakehouse id is provided, the current lakehouse is used.
    """
    if lakehouse == None:
        lakehouse_id = resolve_lakehouse_id()
    elif not _is_valid_uuid(lakehouse):
        lakehouse_name=lakehouse
        return lakehouse_name
    else:
        lakehouse_id = lakehouse

    df_items = fabric.list_items("Lakehouse")
    df_items = df_items[df_items['Id'] == lakehouse_id]
    lakehouse_name = df_items['Display Name'].iloc[0]

    return lakehouse_name


def get_lakehouse_tables_info(spark:SparkSession = None, lakehouse:str | UUID | None = None, workspace:str | UUID | None = None) -> pd.DataFrame:
    """
    Retrieves the table names and related infos from the lakehouse.
    Note: the faster method is to use the sempy API. However, due to a presumably temporary 
          pagination error, it is limited to 100 tables.
          If the number of tables equals to exactly 100, the method falls back to the spark API.
          This code should be modified (by removing the first part of the if condition) once the issue is resolved.

    If workspace_id is not provided, the current workspace is used
    If lakehouse_id is not provided, the current lakehouse is used
    """

    workspace_id = resolve_workspace_id(workspace)
    workspace_name = resolve_workspace_name(workspace)
    lakehouse_id = resolve_lakehouse_id(lakehouse)
    lakehouse_name = resolve_lakehouse_name(lakehouse)

    client = fabric.FabricRestClient()
    response = client.get(f"/v1/workspaces/{workspace_id}/lakehouses/{lakehouse_id}/tables")
    table_list = response.json()['data']
    if len(table_list)==100:
        warn_msg="Warning: The sempy API is truncated to show 100 tables."
        warn_msg+="\nFailing back to the spark API to retrieve all tables."
        warn_msg+="\nThis method is significantly slower."
        print(warn_msg)
        if not spark:
            spark = get_or_create_spark()
        tables = spark.catalog.listTables()
        table_list_formatted = [{"Workspace Name": workspace_name, 
                                 "Workspace ID": workspace_id,
                                 "Lakehouse Name": lakehouse_name, 
                                 "Lakehouse ID": lakehouse_id,
                                 "Table Name": table.name, 
                                 "Type":table.tableType, 
                                 "Location":None, 
                                 "Format":None, 
                                 "Is Temporary": table.isTemporary, 
                                 "Description": table.description} for table in tables]
    else:
        table_list_formatted = [{"Workspace Name": workspace_name, 
                                 "Workspace ID": workspace_id,
                                 "Lakehouse Name": lakehouse_name, 
                                 "Lakehouse ID": lakehouse_id,
                                 "Table Name": table['name'], 
                                 "Type": table['type'], 
                                 "Location": table['location'], 
                                 "Format": table['format'],
                                 "Is Temporary": None, 
                                 "Description": None} for table in table_list]
        
    return pd.DataFrame(table_list_formatted)

if __name__ == "__main__":
    path="abfss://Decide_DK_DEV@onelake.dfs.fabric.microsoft.com/Decide_DK_Bronze.Lakehouse/Files/Bronze/"
    workspace="DEV"
    resolve_workspace_name(None)
    lakehouse=get_lakehouse_from_path(path)
    file_path=get_file_path(workspace=workspace, 
                            lakehouse=lakehouse, 
                            directory="FO_file1")

    print('Done')